package metrics.metric;

import java.util.Optional;
import java.util.function.Function;

import io.vavr.Tuple2;
import io.vavr.Tuple6;
import io.vavr.collection.Array;
import io.vavr.collection.IndexedSeq;
import io.vavr.collection.Stream;
import metrics.structures.CostAndInstedGI;
import metrics.structures.MetricError;
import metrics.structures.JaroValuesGI;
import metrics.structures.MatrixGI;

public class JaroGIInfo<S> implements StructureMatrixGI<JaroValuesGI, S> {
	
	private final IndexedSeq<S> from;
	private final IndexedSeq<S> to;
	private final IndexedSeq<S> finalFrom;
	private final IndexedSeq<S> finalTo;
	private final String output;
	private final String operations;
	private final MatrixGI<JaroValuesGI> mat;
	private final CostAndInstedGI<S> ci;
	private final IndexedSeq<Tuple2<Integer, Integer>> indexes;
	private final Double distance; 
	private final Double weightDistance;
	private final MetricError error;

	private final Function<String, String> translator;
	
	private final double w1;
	private final double w2;
	private final double wt;
	
	public JaroGIInfo(final IndexedSeq<S> from, final IndexedSeq<S> to, final CostAndInstedGI<S> ci,
			 final Function<String, String> translator) {
		
			if( (from == null || to == null) ){
				this.w1=0;
				this.w2 =0;
				this.wt = 0;
				this.from = null;
				this.to = null;
				this.ci = null;
				this.mat = null;
				this.finalFrom = null;
				this.finalTo = null;
				this.output = null;
				this.operations = null;
				this.indexes = null;
				this.distance = null;
				this.weightDistance = null;
				this.translator = null;
				this.error = MetricError.NULL_DATA;
			}else if((from.length()<=0 && to.length()<0)){
				this.w1=0;
				this.w2 =0;
				this.wt = 0;
				this.from = null;
				this.to = null;
				this.ci = null;
				this.mat = null;
				this.finalFrom = null;
				this.finalTo = null;
				this.output = null;
				this.operations = null;
				this.indexes = null;
				this.distance = null;
				this.weightDistance = null;
				this.translator = null;
				this.error = MetricError.WRONG_DATA;
			}else{
				this.w1 = 1.0/3;
				this.w2 = 1.0/3;
				this.wt = 1.0/3;
				
				this.from = from;
				this.to = to;
				this.ci = ci;

				this.translator = translator;
				
				this.mat = fillMatrix(from, to, 
								calculateRange(from, to));
				
				
				
				final Tuple6<IndexedSeq<S>, IndexedSeq<S>, IndexedSeq<Character>, 
							IndexedSeq<Tuple2<Integer, Integer>>, Integer, Integer> 
				finalWords	= findOperations();
				
				
				this.distance = calculateDistance(from, to, finalWords._5, finalWords._6);
				this.weightDistance = calculateDistance(from, to, finalWords._5, finalWords._6);
				
				
				this.finalFrom = finalWords._1;
				this.finalTo = finalWords._2;
				this.operations = operations(finalWords._3);
				this.output = createText(finalWords._5, finalWords._6, operations);
				
				this.indexes = finalWords._4; 
				this.error = MetricError.OK;
			}
	}
	
	public JaroGIInfo(final IndexedSeq<S> from, final IndexedSeq<S> to, final CostAndInstedGI<S> ci,
			final double w1, final double w2, final double wt,
			 final Function<String, String> translator) {
		if( (from == null || to == null) ){
			this.w1=0;
			this.w2 =0;
			this.wt = 0;
			this.from = null;
			this.to = null;
			this.ci = null;
			this.mat = null;
			this.finalFrom = null;
			this.finalTo = null;
			this.output = null;
			this.operations = null;
			this.indexes = null;
			this.distance = null;
			this.weightDistance = null;
			this.translator = null;
			this.error = MetricError.NULL_DATA;
		}else if((from.length()<=0 && to.length()<0)){
			this.w1=0;
			this.w2 =0;
			this.wt = 0;
			this.from = null;
			this.to = null;
			this.ci = null;
			this.mat = null;
			this.finalFrom = null;
			this.finalTo = null;
			this.output = null;
			this.operations = null;
			this.indexes = null;
			this.distance = null;
			this.weightDistance = null;
			this.translator = null;
			this.error = MetricError.WRONG_DATA;
		}else{
			this.w1 = w1;
			this.w2 = w2;
			this.wt = wt;
			
			this.from = from;
			this.to = to;
			this.ci = ci;
			
			this.translator = translator;
			
			this.mat = fillMatrix(from, to, 
							calculateRange(from, to));
			
			final Tuple6<IndexedSeq<S>, IndexedSeq<S>, IndexedSeq<Character>, 
			IndexedSeq<Tuple2<Integer, Integer>>, Integer, Integer> 
			finalWords	= findOperations();
			
			
			this.distance = calculateDistance(from, to, finalWords._5, finalWords._6);
			this.weightDistance = calculateDistance(from, to, finalWords._5, finalWords._6);

			
			this.finalFrom = finalWords._1;
			this.finalTo = finalWords._2;
			this.operations = operations(finalWords._3);
			this.output = createText(finalWords._5, finalWords._6, operations);
			this.indexes = finalWords._4; 
			this.error = MetricError.OK;
		}
	}
	
	
	
	/********************************************/
	
	@Override
	public Optional<IndexedSeq<S>> getFrom() {
		if(from == null)
			return Optional.empty();
		return Optional.of(from);
	}

	@Override
	public Optional<IndexedSeq<S>> getTo() {
		if(to == null)
			return Optional.empty();
		return Optional.of(to);
	}

	@Override
	public Optional<IndexedSeq<S>> getFinalFrom() {
		if(finalFrom == null)
			return Optional.empty();
		return Optional.of(finalFrom);
	}

	@Override
	public Optional<IndexedSeq<S>> getFinalTo() {
		if(finalTo == null)
			return Optional.empty();
		return Optional.of(finalTo);
	}

	@Override
	public Optional<MatrixGI<JaroValuesGI>> getStructure() {
		if(mat == null)
			return Optional.empty();
		return Optional.of(mat);
	}

	@Override
	public Optional<Number> getDistance() {
		if(distance == null)
			return Optional.empty();
		return Optional.of(distance);
	}

	@Override
	public Optional<Number> getWeightDistance() {
		if(this.weightDistance == null)
			return Optional.empty();
		return Optional.of(weightDistance);
	}
	
	@Override
	public Optional<String> getFormatedText() {
		if(output == null)
			return Optional.empty();
		return Optional.of(output);
	}


	@Override
	public Optional<IndexedSeq<Tuple2<Integer, Integer>>> getIndexes() {
		if(indexes == null)
			return Optional.empty();
		return Optional.of(indexes);
	}


	@Override
	public int getShift() {
		return 0;
	}


	@Override
	public String getValue(final JaroValuesGI origin) {
		switch (origin) {
		case TRUE:return "1";
		case FALSE: return "0";
		case NULL: return "-";
		default: return "-";
		}
	}

	@Override
	public MetricError getError() {
		return error;
	}
	
	public String getOperations(){
		return operations;
	}
	
	/******************************************/
	
	/**
	 * metoda vytvory matici a vyplni ji hodnotami
	 * @param from
	 * @param to
	 * @param r - rozsah vypoctu
	 * @return
	 */
	private MatrixGI<JaroValuesGI> fillMatrix(final IndexedSeq<S> from, final IndexedSeq<S> to, final double r){
		final JaroValuesGI[][] aux= new JaroValuesGI[from.length()][to.length()];
		final MatrixGI<JaroValuesGI> result = new MatrixGI<>(aux);
		return Stream.range(0, result.getRowSize()).foldLeft(result,
				(row, i ) ->Stream.range(0, result.getColumnSize()).foldLeft(result,
						(column, j)->{
							if(Math.abs(i-j) < r && from.get(i).equals( to.get(j) ) )
								return result.withCell(i, j, JaroValuesGI.TRUE);
							if(Math.abs(i-j) < r )
								return result.withCell(i, j, JaroValuesGI.FALSE);
							return result.withCell(i, j, JaroValuesGI.NULL);
						} ) );
	}
	
	
	
	/**
	 * projde matici a vytvory final from a to
	 * string podobnosti pro popis
	 * seznam indexu pro zobrazeni
	 * @return
	 */
	private Tuple6<IndexedSeq<S>, IndexedSeq<S>,
					IndexedSeq<Character>,IndexedSeq<Tuple2<Integer,Integer>>,Integer,Integer> 
				findOperations(){
		return quadrates(-1, -1, new Tuple6<IndexedSeq<S>, IndexedSeq<S>, IndexedSeq<Character>, 
						 IndexedSeq<Tuple2<Integer, Integer>>, Integer, Integer>(
								 Array.empty(), Array.empty(), Array.empty(), Array.empty(), 0, 0));
	} 
	
	private Tuple6<IndexedSeq<S>, IndexedSeq<S>,IndexedSeq<Character>,
					IndexedSeq<Tuple2<Integer,Integer>>,Integer,Integer>  quadrates(
			final int n, final int m,
			final Tuple6<
						IndexedSeq<S>, 
						IndexedSeq<S>,
						IndexedSeq<Character>,
						IndexedSeq<Tuple2<Integer,Integer>>,
						Integer,
						Integer
					   >  result){
		//v pravem dolnim rohu
	//	if(n+1>=mat.getRowSize() && m+1>=mat.getColumnSize())
		if(n+1>=from.length() && m+1>=to.length())
			return result;
		//hrany
	//	if(n+1>=mat.getRowSize()) //smazani
		if(n+1>=from.length()) //smazani
			return quadrates(n, m+1, 
					new Tuple6<IndexedSeq<S>, IndexedSeq<S>, IndexedSeq<Character>, 
					 IndexedSeq<Tuple2<Integer, Integer>>, Integer, Integer>(
							result._1.append(ci.getInsertion().apply(' ')),
							result._2.append(to.get(m+1)),
							result._3.append('I'),
							result._4.append(new Tuple2<>(n, m+1)),
							result._5,
							result._6
					));
	//	if(m+1>= mat.getColumnSize())//vlozeni
		if(m+1>= to.length())//vlozeni
			return quadrates(n+1, m, 
					new Tuple6<IndexedSeq<S>, IndexedSeq<S>, IndexedSeq<Character>, 
					 IndexedSeq<Tuple2<Integer, Integer>>, Integer, Integer>(
							result._1.append(from.get(n+1)),
							result._2.append(ci.getDeletion().apply(' ')),
							result._3.append('D'),
							result._4.append(new Tuple2<>(n+1, m)),
							result._5,
							result._6
					));

		/***********************************/
	/*	System.out.println(n + " " + m);
		System.out.println("bude stejna? " + mat.get(n+1, m+1));
		System.out.println();
		*/
		//stejne
		if(mat.get(n+1, m+1)==JaroValuesGI.TRUE)
			return quadrates(n+1, m+1, 
					new Tuple6<IndexedSeq<S>, IndexedSeq<S>, IndexedSeq<Character>, 
					 IndexedSeq<Tuple2<Integer, Integer>>, Integer, Integer>(
							result._1.append(from.get(n+1)),
							result._2.append(to.get(m+1)),
							result._3.append('E'),
							result._4.append(new Tuple2<>(n+1, m+1)),
							result._5 + 1,
							result._6
					));
		//hodnota n+1, m+1 je vzdy false			
		//transpozice
		if(isTransposition(n, m))
			return quadrates(n+2, m+2, 
					new Tuple6<IndexedSeq<S>, IndexedSeq<S>, IndexedSeq<Character>, 
					 IndexedSeq<Tuple2<Integer, Integer>>, Integer, Integer>(
							result._1.append(ci.getTransposition().apply( from.get(n+1)))
									 .append(ci.getTransposition().apply(from.get(n+2))) ,
							result._2.append(ci.getTransposition().apply(to.get(m+1)))
									 .append(ci.getTransposition().apply(to.get(m+2))),
							result._3.append('T').append('T'),
							result._4.append(new Tuple2<>(n+2, m+1))
							         .append(new Tuple2<Integer, Integer>(n+1, m+2)),
							result._5 + 2,
							result._6 + 1
									 //+ 2
					));
		final int direct = look(n+1, m+1);
		int match = 0;
		//smazani
		if(direct == 0){
			if(n >= 0 && m >= 0)
				if(mat.get(n, m) == JaroValuesGI.FALSE && mat.get(n+1, m)==JaroValuesGI.TRUE)
					match = 1;
			return quadrates(n+1, m, 
					new Tuple6<IndexedSeq<S>, IndexedSeq<S>, IndexedSeq<Character>, 
					 IndexedSeq<Tuple2<Integer, Integer>>, Integer, Integer>(
							result._1.append(from.get(m+1)),
							result._2.append(ci.getDeletion().apply(' ')),
							result._3.append('D'),
							result._4.append(new Tuple2<>(n+1, m)),
							result._5 + match,
							result._6
					));
		}
		//vlozeni
		if(direct == 1){
			if(n >= 0 && m >= 0)
				if(mat.get(n, m) == JaroValuesGI.FALSE && mat.get(n, m+1)==JaroValuesGI.TRUE)
					match = 1;
			return quadrates(n, m+1, 
					new Tuple6<IndexedSeq<S>, IndexedSeq<S>, IndexedSeq<Character>, 
					 IndexedSeq<Tuple2<Integer, Integer>>, Integer, Integer>(
							result._1.append(ci.getInsertion().apply(' ')), 
							result._2.append(to.get(n+1)),
							result._3.append('I'),
							result._4.append(new Tuple2<>(n, m+1)),
							result._5 + match,
							result._6
					));		
		}
		//vlozeni a smazani - substituce
		return quadrates(n+1, m+1, 
				new Tuple6<IndexedSeq<S>, IndexedSeq<S>, IndexedSeq<Character>, 
				 IndexedSeq<Tuple2<Integer, Integer>>, Integer, Integer>(
						result._1.append(ci.getInsertion().apply(' ')).append(from.get(n+1)),
						result._2.append(to.get(m+1)).append(ci.getDeletion().apply(' ')),
						result._3.append('I').append('D'),
						result._4.append(new Tuple2<>(n+1, m))
						         .append(new Tuple2<Integer, Integer>(n+1, m+1)),
						result._5,
						result._6
				));
	}
	
	
	
	
	
	
	/**
	 * pocita rozsah vypoctu v matici  - r
	 * @param from
	 * @param to
	 * @return
	 */
	private double calculateRange(final IndexedSeq<S> from, final IndexedSeq<S> to){
		if(Math.max(from.length(), to.length())/2.0 - 1 > 1)
			return Math.max(from.length(), to.length())/2.0 - 1;
		return Math.min(from.length(), to.length())/2.0;
		//return Math.min(from.length(), to.length())/2.0;
		//return Math.max(from.length(), to.length())/2.0 - 1;
	}
	
	/**
	 * spocita vzdalenost pomoci vzorce
	 * @param c pocet shod
	 * @param t pocet transpozic
	 * @return
	 */
	private double calculateDistance(final IndexedSeq<S> from, final IndexedSeq<S> to,
									 final int c, final int t){
		if(from.length()==0)
			return 0.0;
		if(to.length()==0)
			return 0.0;
		if(c==0)
			return 0.0;
		return (w1*c/(double)from.length()) + (w2*c/(double)to.length()) + ( wt*(c-t)/(double)c ) ;
	}
	

	/**
	 * vytvori slovni popis
	 * @return
	 */
	private String createText(final int matches, final int trans, final String oper) { //final Tuple2<Integer, Integer> ct
		return  oper
				+ "\n"
				+ this.translator.apply("formula") 
				+ ": W1c/|p| + W2c/|d| + Wt(c-t)/c \n"
				+ this.translator.apply("instate")
				+ ": "
						+ w1+""+ matches+"/"+ from.length()+"\n\t + "
						+ w2+""+ matches+"/"+ to.length() + "\n\t + "
						+ wt+"("+matches+"-"+trans+")/"+matches;
	}
	
	
	
	/**
	 * popis operaci
	 * je zvlast pro jaro-winklerovu vzdalenost
	 * @param sim
	 * @return
	 */
	private String operations(final IndexedSeq<Character> sim){
		return  Stream.range(0, sim.length()).foldLeft("",
				(curr, i)->{
					if(sim.get(i)=='I')
						return curr
								+ this.translator.apply("insertion") 
								+ " '"
								+ finalTo.get(i)
								+ "' "
								+ this.translator.apply("position") 
								+ " "
								+ (i+1)+"\n";
					if(sim.get(i)=='D')
						return curr
								+ this.translator.apply("deletion") 
								+ " '"
								+ finalFrom.get(i)
								+ "' "
								+ this.translator.apply("position") 
								+ " "
								+ (i+1)+"\n";
					if(sim.get(i)=='T' && i>0  && sim.get(i-1)!='T' )
						return curr
								+ this.translator.apply("transposition") 
								+ " '"
								+ finalFrom.get(i)
								+ "' "
								+ this.translator.apply("and") 
								+ " '"
								+ finalFrom.get(i+1)
								+ "' "
								+ this.translator.apply("position") 
								+ " "
								+ (i+1)
								+ " "
								+ this.translator.apply("and") 
								+ " "
								+ (i+2)
								+ "\n";
					return curr;
				});
	}
	
	
	
	/**
	 * projde radek/sloupec/diagonalu a hleda prvek s cost 0
	 * @param i
	 * @param j
	 * @return 0 jit ve sloupci, 1 jit v radku, ostatni diagonala
	 */
	private int look(final int i,final int j){
		final double range = calculateRange(from, to);
		final int dia = distance(i, j, 2, range);
		final int col = distance(i, j, 1, range);
		final int row = distance(i, j, 0, range);
		
		if(col != -1 && check(col, dia, true) && check(col, row, true))//smazani
			return 0;
		if(row != -1 && check(row, dia, true) && check(row, col, false))//vlozeni
			return 1;
		return -1;
	}
	
	private boolean check(final int expected, final int actual, final boolean equal){
		if(actual == -1)
			return true;
		if(equal)
			return expected <= actual;
		return expected < actual;
	}
	
	
	/**
	 * projde radek/sloupec/diagonalu od zadaneho bodu a hleda bunku s hodnotou true
	 * @param i
	 * @param j
	 * @param mode - 0 radek, 1 sloupec, ostatni - diagonala
	 * @return vzdalenost k nejblizsimu, nebo rozsah zkoumani
	 */
	private int distance(final int i, final int j, final int mode, final double range){
		//final int range = Math.min(from.length(), to.length())/2;
		if(mode == 0) //radek
			return Stream.range(j, mat.getColumnSize()).foldLeft(-1, 
					(curr, n)->{
						if(curr!=-1 && curr < j+range) //mat.getColumnSize()
							return curr;
						if(getBoolean(i, n))
							return n-j;
						return curr;
					});
		if(mode == 1) //sloupec
			return Stream.range(i, mat.getRowSize()).foldLeft(-1 , 
					(curr, n)->{
						if(curr!=-1 && curr < i+range) //mat.getRowSize()  
							return curr;
						if(getBoolean(n, j))
							return n-i;
						return curr;
					});	
		//diagonala
		return Stream.range(0, mat.getRowSize()).foldLeft(-1, 
				(curr, n)->{
					if(curr!=-1 && curr < range) //mat.getRowSize()
						return curr;
					if( i+n< mat.getRowSize() && j+n<mat.getColumnSize() &&
							getBoolean(i+n, j+n))
						return n;
					return curr;
				});
	}
	
	
	
	/**
	 * zjisti, zda dalsi ctverec je transpozice
	 * sama zkontroluje, zda n m neni mimo matici
	 * @param n - aktualni radek
	 * @param m - aktualni sloupec
	 * @return
	 */
	private boolean isTransposition(final int n, final int m){
		if(n+2<mat.getRowSize() && m+2 <mat.getColumnSize()
		   && mat.get(n+2, m+1)==JaroValuesGI.TRUE && mat.get(n+1, m+2)==JaroValuesGI.TRUE	)
			return true;
		return false;
	}

	/**
	 * zda je hodnota bunky TRUE
	 * @param i
	 * @param j
	 * @return
	 */
	private boolean getBoolean(final int i, final int j){
		if(mat.get(i, j)==JaroValuesGI.TRUE )
			return true;
		return false;
	}


}
