package metrics.metric;

import java.util.Optional;
import java.util.function.Function;

import io.vavr.Tuple2;
import io.vavr.collection.IndexedSeq;
import io.vavr.collection.Stream;
import metrics.structures.CostAndInstedGI;
import metrics.structures.MetricError;
import metrics.structures.JaroValuesGI;
import metrics.structures.MatrixGI;

public class JaroWinklerGIInfo<S> implements StructureMatrixGI<JaroValuesGI, S> {
	
	private final JaroGIInfo<S> jaro;
	private final Double distance;
	private final double p;
	private final double alfa;
	private final String output; 
	private final MetricError error;

	private final Function<String, String> translator;
	
	public JaroWinklerGIInfo(final IndexedSeq<S> from, final IndexedSeq<S> to, final CostAndInstedGI<S> ci,
			final Function<String, String> translator) {
		this.jaro = new JaroGIInfo<S>(from, to, ci, translator);
		if( (from == null || to == null)){
			this.distance = null;
			this.p = 0;
			this.alfa = 0;
			this.output = null;
			this.translator = null;
			this.error = MetricError.NULL_DATA;
		}else if((from.length()<=0 && to.length()<0)	){
			this.distance = null;
			this.p = 0;
			this.alfa = 0;
			this.output = null;
			this.translator = null;
			this.error = MetricError.WRONG_DATA;
		}else{
			this.p = 0.1;
			this.alfa = getAlfa(from, to);
			this.translator = translator;
			this.distance = calculateDistance(p, alfa);
			this.output = createText(jaro);
			this.error = MetricError.OK;
		}
	}
	
	
	public JaroWinklerGIInfo(final IndexedSeq<S> from, final IndexedSeq<S> to,final CostAndInstedGI<S> ci, final double p,
			final Function<String, String> translator) {
		this.jaro = new JaroGIInfo<S>(from, to, ci, translator);
		if( (from == null || to == null)){
			this.distance = null;
			this.p = 0;
			this.alfa = 0;
			this.output = null;
			this.translator = null;
			this.error = MetricError.NULL_DATA;
		}else if((from.length()<=0 && to.length()<0)	){
			this.distance = null;
			this.p = 0;
			this.alfa = 0;
			this.output = null;
			this.translator = null;
			this.error = MetricError.WRONG_DATA;
		}else{
			this.p = p;
			this.translator = translator;
			this.alfa = getAlfa(from, to);
			this.distance = calculateDistance(p, alfa);
			this.output = createText(jaro);
			this.error = MetricError.OK;
		}
	}
	
	/***********************************/
	@Override
	public Optional<IndexedSeq<S>> getFrom() {
		return jaro.getFrom();
	}
	@Override
	public Optional<IndexedSeq<S>> getTo() {
		return jaro.getTo();
	}
	@Override
	public Optional<IndexedSeq<S>> getFinalFrom() {
		return jaro.getFinalFrom();
	}
	@Override
	public Optional<IndexedSeq<S>> getFinalTo() {
		return jaro.getFinalTo();
	}
	@Override
	public Optional<MatrixGI<JaroValuesGI>> getStructure() {
		return jaro.getStructure();
	}
	@Override
	public Optional<Number> getDistance() {
		if(distance == null)
			return Optional.empty();
		return Optional.of(distance);
	}
	@Override
	public Optional<String> getFormatedText() {
		if(output == null)
			return Optional.empty();
		return Optional.of(output);
	}
	@Override
	public Optional<IndexedSeq<Tuple2<Integer, Integer>>> getIndexes() {
		return jaro.getIndexes();
	}
	@Override
	public int getShift() {
		return jaro.getShift();
	}
	@Override
	public String getValue(final JaroValuesGI origin) {
		return jaro.getValue(origin);
	}

	@Override
	public MetricError getError() {
		return error;
	}
	@Override
	public Optional<Number> getWeightDistance() {
		return jaro.getWeightDistance();
	}
	
	/**************************************/
	
	/**
	 * vypocita vzdalenost na zaklade jaroovy metriky
	 * @param p
	 * @param alfa
	 * @return
	 */
	private double calculateDistance( final double p,final double alfa) {
		final Number jaroDistance = jaro.getDistance().orElse(0.0);
		return jaroDistance.doubleValue() + (alfa*p*(1-jaroDistance.doubleValue() ));
	}

	/**
	 * zjisteni alfa slozky - velikost spolecneho zakladu do velikosti 4
	 * @param from
	 * @param to
	 * @return
	 */
	private double getAlfa(final IndexedSeq<S> from,final  IndexedSeq<S> to) {
		return Stream.range(0, Math.min(4, Math.min(from.length(), to.length())))
				.foldLeft(0,
					(current, i)->{
						if(i>current)
							return current;
						if(from.get(i).equals( to.get(i) ) )
							return current +1;
						return current;
				});
	}
	
	/**
	 * vytvori slovni popis
	 * @return
	 */
	private String createText(final JaroGIInfo<S> jaro) {
		return jaro.getOperations()
				+ this.translator.apply("formula") 
				+ ": dj + ip(1-dj) \n"
				+ this.translator.apply("instate") 
				+ ": "
				+jaro.getDistance().get() +"  "+ alfa
							  + ""+p+"(1-"+jaro.getDistance().get()+")";
	}


	
}
