package metrics.metricquick;

import java.util.Optional;

import io.vavr.Tuple2;
import io.vavr.collection.IndexedSeq;
import io.vavr.collection.Stream;
import metrics.structures.MetricError;
import metrics.structures.JaroValuesGI;
import metrics.structures.MatrixGI;

public class JaroGIQuick<S> implements MetricGIQuick<S> {

	private final IndexedSeq<S> from;
	private final IndexedSeq<S> to;
	private final Double distance; 
	private final MetricError error;
	
	public JaroGIQuick(final IndexedSeq<S> from, final IndexedSeq<S> to) {
		if( (from == null || to == null) ){
			this.from = null;
			this.to = null;
			this.distance = null;
			this.error = MetricError.NULL_DATA;
		}else if((from.length()<=0 && to.length()<0)){
			this.from = null;
			this.to = null;
			this.distance = null;
			this.error = MetricError.WRONG_DATA;
		}else{
			this.from = from;
			this.to = to;
			
			final Tuple2<Integer, Integer> ct = calculateCT(
													fillMatrix(from, to,
														calculateRange(from, to)));
			this.distance = calculateDistance(
									from.length(), to.length(),
									ct._1, ct._2,
									1.0/3,1.0/3,1.0/3);
			this.error = MetricError.OK;
		}	
		
	}
	
	public JaroGIQuick(final IndexedSeq<S> from, final IndexedSeq<S> to,
			final double w1, final double w2, final double wt) {
		if( (from == null || to == null) ){
			this.from = null;
			this.to = null;
			this.distance = null;
			this.error = MetricError.NULL_DATA;
		}else if((from.length()<=0 && to.length()<0)){
			this.from = null;
			this.to = null;
			this.distance = null;
			this.error = MetricError.WRONG_DATA;
		}else{
			this.from = from;
			this.to = to;
			
					
			final Tuple2<Integer, Integer> ct = calculateCT(
													fillMatrix(from, to,
															calculateRange(from, to)));
			
			this.distance = calculateDistance(
									from.length(), to.length(),
									ct._1, ct._2,
									w1,w2,wt);
			this.error = MetricError.OK;
		}
	}
	
	
	@Override
	public Optional<IndexedSeq<S>> getFrom() {
		if(from == null)
			return Optional.empty();
		return Optional.of(from);
	}

	@Override
	public Optional<IndexedSeq<S>> getTo() {
		if(to == null)
			return Optional.empty();
		return Optional.of(to);
	}

	@Override
	public Optional<Number> getDistance() {
		if(distance == null)
			return Optional.empty();
		return Optional.of(distance);
	}

	@Override
	public MetricError getError() {
		return error;
	}
	
	/**
	 * metoda vytvory matici a vyplni ji hodnotami
	 * @param from
	 * @param to
	 * @param r - rozsah vypoctu
	 * @return
	 */
	private MatrixGI<JaroValuesGI> fillMatrix(final IndexedSeq<S> from, final IndexedSeq<S> to, final double r){
		final JaroValuesGI[][] aux= new JaroValuesGI[from.length()][to.length()];
		final MatrixGI<JaroValuesGI> result = new MatrixGI<>(aux);
		return Stream.range(0, result.getRowSize()).foldLeft(result,
				(row, i ) ->Stream.range(0, result.getColumnSize()).foldLeft(result,
						(column, j)->{
							if(Math.abs(i-j) < r && from.get(i).equals( to.get(j) ) )
								return result.withCell(i, j, JaroValuesGI.TRUE);
							if(Math.abs(i-j) < r )
								return result.withCell(i, j, JaroValuesGI.FALSE);
							return result.withCell(i, j, JaroValuesGI.NULL);
						} ) );
	}
	

	/**
	 * pocita rozsah vypoctu v matici  - r
	 * @param from
	 * @param to
	 * @return
	 */
	private double calculateRange(final IndexedSeq<S> from, final IndexedSeq<S> to){
		if(Math.max(from.length(), to.length())/2.0 - 1 > 2)
			return Math.max(from.length(), to.length())/2.0 - 1;
		return Math.min(from.length(), to.length())/2.0;
		//return Math.min(from.length(), to.length())/2.0;
		//return Math.max(from.length(), to.length())/2.0 - 1;
	}
	
	/**
	 * spocita vzdalenost pomoci vzorce
	 * @param c pocet shod
	 * @param t pocet transpozic
	 * @return
	 */
	private double calculateDistance(
			final int fromLength, final int toLength,
									 final int c, final int t,
									 final double w1, final double w2, final double wt){
		if(fromLength==0)
			return 0.0;
		if(toLength==0)
			return 0.0;
		if(c==0)
			return 0.0;
		return (w1*c/(double)from.length()) + (w2*c/(double)to.length()) + ( wt*(c-t)/(double)c ) ;
	}
	
	
	/**
	 * projde matici a vytvory final from a to
	 * string podobnosti pro popis
	 * seznam indexu pro zobrazeni
	 * @return
	 */
	private Tuple2<Integer,Integer> 
				calculateCT(final MatrixGI<JaroValuesGI> mat){
		return quadrates(-1, -1, new Tuple2<Integer, Integer>(0, 0), mat);
	} 
	
	private Tuple2<Integer,Integer>  quadrates(
			final int n, final int m,
			final Tuple2<Integer,Integer>  result, final MatrixGI<JaroValuesGI> mat){
		//na hrane
	//	if(n+1>=mat.getRowSize() && m+1>=mat.getColumnSize())
		if(n+1>=from.length() || m+1>=to.length())
			return result;
		/***********************************/
		//stejne
		if(mat.get(n+1, m+1)==JaroValuesGI.TRUE)
			return quadrates(n+1, m+1, 
					new Tuple2<Integer,Integer>(result._1 + 1,result._2), mat);
		//hodnota n+1, m+1 je vzdy false		
		//transpozice
		if(isTransposition(n, m, mat))
			return quadrates(n+2, m+2, 
					new Tuple2<Integer,Integer>(result._1 + 2,result._2 + 1), mat);
		final int direct = look(n+1, m+1, mat);
		int match = 0;
		//smazani
		if(direct == 0){
			if(n >= 0 && m >= 0)
				if(mat.get(n, m) == JaroValuesGI.FALSE && mat.get(n+1, m)==JaroValuesGI.TRUE)
					match = 1;
			return quadrates(n+1, m, 
					new Tuple2<Integer,Integer>(result._1 + match ,result._2), mat);
		}	
		//vlozeni
		if(direct == 1){
			if(n >= 0 && m >= 0)
				if(mat.get(n, m) == JaroValuesGI.FALSE && mat.get(n, m+1)==JaroValuesGI.TRUE)
					match = 1;
			return quadrates(n, m+1, 
					new Tuple2<Integer,Integer>(result._1 + match ,result._2), mat);
		}	
		//vlozeni a smazani - substituce
		return quadrates(n+1, m+1, 
				new Tuple2<Integer,Integer>(result._1,result._2), mat);
	}
	
	

	/**
	 * projde radek/sloupec/diagonalu a hleda prvek s cost 0
	 * @param i
	 * @param j
	 * @return 0 jit ve sloupci, 1 jit v radku, ostatni diagonala
	 */
	private int look(final int i,final int j,
			final MatrixGI<JaroValuesGI> mat){
		final double range = calculateRange(from, to);
		final int dia = distance(i, j, 2, range, mat);
		final int col = distance(i, j, 1, range, mat);
		final int row = distance(i, j, 0, range, mat);
		
		if(col != -1 && check(col, dia, true) && check(col, row, true))//smazani
			return 0;
		if(row != -1 && check(row, dia, true) && check(row, col, false))//vlozeni
			return 1;
		return -1;
	}
	
	private boolean check(final int expected, final int actual, final boolean equal){
		if(actual == -1)
			return true;
		if(equal)
			return expected <= actual;
		return expected < actual;
	}
	
	
	/**
	 * projde radek/sloupec/diagonalu od zadaneho bodu a hleda bunku s hodnotou true
	 * @param i
	 * @param j
	 * @param mode - 0 radek, 1 sloupec, ostatni - diagonala
	 * @return vzdalenost k nejblizsimu, nebo rozsah zkoumani
	 */
	private int distance(final int i, final int j, final int mode, final double range,
			final MatrixGI<JaroValuesGI> mat){
		//final int range = Math.min(from.length(), to.length())/2;
		if(mode == 0) //radek
			return Stream.range(j, mat.getColumnSize()).foldLeft(-1, 
					(curr, n)->{
						if(curr!=-1 && curr < j+range) //mat.getColumnSize()
							return curr;
						if(getBoolean(i, n, mat))
							return n-j;
						return curr;
					});
		if(mode == 1) //sloupec
			return Stream.range(i, mat.getRowSize()).foldLeft(-1 , 
					(curr, n)->{
						if(curr!=-1 && curr < i+range) //mat.getRowSize()  
							return curr;
						if(getBoolean(n, j, mat))
							return n-i;
						return curr;
					});	
		//diagonala
		return Stream.range(0, mat.getRowSize()).foldLeft(-1, 
				(curr, n)->{
					if(curr!=-1 && curr < range) //mat.getRowSize()
						return curr;
					if( i+n< mat.getRowSize() && j+n<mat.getColumnSize() &&
							getBoolean(i+n, j+n , mat))
						return n;
					return curr;
				});
	}
	
	
	
	/**
	 * zjisti, zda dalsi ctverec je transpozice
	 * sama zkontroluje, zda n m neni mimo matici
	 * @param n - aktualni radek
	 * @param m - aktualni sloupec
	 * @return
	 */
	private boolean isTransposition(final int n, final int m, final MatrixGI<JaroValuesGI> mat){
		if(n+2<mat.getRowSize() && m+2 <mat.getColumnSize()
		   && mat.get(n+2, m+1)==JaroValuesGI.TRUE && mat.get(n+1, m+2)==JaroValuesGI.TRUE	)
			return true;
		return false;
	}

	/**
	 * zda je hodnota bunky TRUE
	 * @param i
	 * @param j
	 * @return
	 */
	private boolean getBoolean(final int i, final int j, final MatrixGI<JaroValuesGI> mat){
		if(mat.get(i, j)==JaroValuesGI.TRUE )
			return true;
		return false;
	}


	
	
	

}
